@@ -11,10 +11,9 @@ It also defines a couple of simple job classes.
 import sys
 import asyncio
 
-from .dotstyle import DotStyle
+from .bestset import BestSet
 
-debug = False                                           # pylint: disable=C0103
-debug = True                                            # pylint: disable=C0103
+from .dotstyle import DotStyle
 
 # pylint settings
 # W0212: we have a lot of accesses to protected members of other classes
@@ -110,7 +109,7 @@ class AbstractJob:                                      # pylint: disable=R0902
     """
 
     def __init__(self, *,                               # pylint: disable=R0913
-                 forever=False, critical=False, label=None,
+                 forever=False, critical=True, label=None,
                  required=None, scheduler=None):
         self.forever = forever
         self.critical = critical
@@ -118,7 +117,7 @@ class AbstractJob:                                      # pylint: disable=R0902
         self.label = label
         # for convenience, one can mention
         # only one, or a collection of, AbstractJobs
-        self.required = set()
+        self.required = BestSet()
         self.requires(required)
         # convenience again
         if scheduler is not None:
@@ -133,7 +132,7 @@ class AbstractJob:                                      # pylint: disable=R0902
         # this is for graph browsing algos
         self._s_mark = None
         # the reverse of required
-        self._s_successors = set()
+        self._s_successors = BestSet()
         # this attribute is reserved for use by the scheduler
         # that will for example store there an ordering information,
         # which in turn can be used for printing relationships
@@ -161,21 +160,38 @@ class AbstractJob:                                      # pylint: disable=R0902
         """
         return 1
 
-    # we don't use depth
-    def _list(self, details, _):
+    # don't use parameter recursive
+    def _list(self, details, depth, _):
         """
-        Complicit to Scheduler.list()
+        Complicit to PureScheduler.list()
         """
-        print(self.repr_id(),
-              self.repr_short(),
-              self.repr_main(),
-              self.repr_result(),
-              self.repr_requires())
+        indent = ('>'*depth + ' ') if depth else ''
+        print("{} {} {}{} {} {}"
+              .format(self.repr_id(),
+                      self.repr_short(),
+                      indent,
+                      self.repr_main(),
+                      self.repr_result(),
+                      self.repr_requires()))
         if details and hasattr(self, 'details'):
-            details = self.details()
+            details = self.details()                    # pylint: disable=e1111
             if details is not None:
                 print(details)
 
+    # don't use parameter recursive
+    def _list_safe(self, _):
+        """
+        Complicit to PureScheduler.list_safe()
+        """
+        print("{} {} {} {}"
+              .format(self.repr_short(),
+                      self.repr_id(),
+                      self.repr_main(),
+                      self.repr_requires()))
+
+    def _iterate_jobs(self, scan_schedulers):           # pylint: disable=w0613
+        yield self
+
     def _get_text_label(self):
         # In terms of labelling, things have become a little tricky over
         # time. When listing an instance of Scheduler, there are 2 ways
@@ -194,7 +210,7 @@ class AbstractJob:                                      # pylint: disable=R0902
         if attempt is not None:
             return attempt
         # otherwise, try self.text_label() and use that
-        attempt = self.text_label()
+        attempt = self.text_label()                     # pylint: disable=e1111
         if attempt is not None:
             return attempt
         # otherwise
@@ -205,10 +221,10 @@ class AbstractJob:                                      # pylint: disable=R0902
         # we don't need to bother about _sched_id here though
         # also we try graph_label() first, and resort to text_label()
         # if it's not redefined on the object
-        attempt = self.graph_label()
+        attempt = self.graph_label()                    # pylint: disable=e1111
         if attempt is not None:
             return attempt
-        return self._get_text_label()
+        return "{}: {}".format(self.repr_id(), self._get_text_label())
 
     def text_label(self):
         """
@@ -254,31 +270,39 @@ class AbstractJob:                                      # pylint: disable=R0902
 
         Legend is quite simply that:
 
-        * critical jobs have a red border, and
-        * the border of forever jobs is made of a dashed line.
+        * schedulers have sharp angles, while other jobs have
+          rounded corners,
+        * critical jobs have a colored and thick border, and
+        * forever jobs have a dashed border.
 
         Returns:
           DotStyle: a dict-like mapping that sets DOT
           attributes for that job.
         """
         # see also https://graphviz.gitlab.io/_pages/doc/info/attrs.html
-        data = DotStyle()
+        style = DotStyle()
         # style; DotStyle known how to deal with lists
-        data['style'] = []
+        style['style'] = []
         # label
         label = self._get_graph_label()
         if label != 'NOLABEL':
-            data['label'] = label
+            style['label'] = label
         # common attributes : rounded box
-        data['shape'] = 'box'
-        data['style'].append('rounded')
+        style['shape'] = 'box'
+        # schedulers keep sharp corners, jobs have rounded corners
+        from .purescheduler import PureScheduler
+        if not isinstance(self, PureScheduler):
+            style['style'].append('rounded')
         # forever items have a dashed line
         if self.forever:
-            data['style'].append('dashed')
-        # critical elements have a red border
+            style['style'].append('dashed')
+        # critical elements have a thicker red border
         if self.is_critical():
-            data['color'] = 'red'
-        return data
+            style['color'] = 'red'
+            style['penwidth'] = 2
+        else:
+            style['penwidth'] = 0.5
+        return style
 
     ##########
     _has_support_for_unicode = None  # type: bool
@@ -287,7 +311,9 @@ class AbstractJob:                                      # pylint: disable=R0902
     def _detect_support_for_unicode(cls):
         if cls._has_support_for_unicode is None:
             try:
-                cls._c_saltire.encode(sys.stdout.encoding)
+                # somehow on the test VB ubuntu box for apssh
+                # sys.stdout.encoding is None
+                cls._c_saltire.encode(sys.stdout.encoding or 'UTF-8')
                 cls._has_support_for_unicode = True
             except UnicodeEncodeError:
                 cls._has_support_for_unicode = False
@@ -430,11 +456,15 @@ class AbstractJob:                                      # pylint: disable=R0902
                     self.repr_result(),
                     self.repr_requires()))
 
-    def requires(self, *requirements):
+    def requires(self, *requirements, remove=False):
         """
         Arguments:
           requirements: an iterable of `AbstractJob`
             instances that are added to the requirements.
+          remove(bool): if set, the requirement are dropped rather than added.
+
+        Raises:
+          KeyError: when trying to remove dependencies that were not present.
 
         For convenience, any nested structure made of job instances
         can be provided, and if None objects are found, they are silently
@@ -450,23 +480,31 @@ class AbstractJob:                                      # pylint: disable=R0902
         * ``j1.requires(j2, [j3, j4])``
         * ``j1.requires((j2, j3))``
         * ``j1.requires(([j2], [[[j3]]]))``
+
+        * Any of the above with ``remove=True``.
+
+        For dropping dependencies instead of adding them, use ``remove=True``
+
         """
         from .sequence import Sequence
         for requirement in requirements:
             if requirement is None:
                 continue
             if isinstance(requirement, AbstractJob):
-                self.required.add(requirement)
+                if not remove:
+                    self.required.add(requirement)
+                else:
+                    self.required.remove(requirement)
             elif isinstance(requirement, Sequence):
                 if requirement.jobs:
                     self.required.add(requirement.jobs[-1])
             elif isinstance(requirement, (tuple, list)):
                 for req in requirement:
-                    self.requires(req)
+                    self.requires(req, remove=remove)
             # not quite sure about what do to here in fact
             else:
                 print("WARNING: fishy requirement in AbstractJob.requires")
-                self.requires(list(requirement))
+                self.requires(list(requirement), remove=remove)
 
     def is_idle(self):
         """
@@ -548,7 +586,7 @@ class AbstractJob:                                      # pylint: disable=R0902
 
     async def co_shutdown(self):
         """
-        Abstract virtual - needs to be implemented
+        Abstract virtual - needs to be implemented.
         """
         print("AbstractJob.co_shutdown() needs to be implemented on class {}"
               .format(self.__class__.__name__))
