@@ -7,9 +7,10 @@ The PureScheduler class is a set of AbstractJobs, that together with their
 
 import time
 import io
-
 import asyncio
 
+from .bestset import BestSet
+
 from .job import AbstractJob
 from .sequence import Sequence
 from .window import Window
@@ -18,15 +19,15 @@ from .dotstyle import DotStyle
 
 #
 # will hopefully go away some day
-debug = False                                           # pylint: disable=C0103
-# debug = True
+DEBUG = False                                           # pylint: disable=C0103
+# DEBUG = True
 
 # pylint settings
 # W0212: we have a lot of accesses to protected members of other classes
 # R0914 Too many local variables
 # R0904 Too many public methods (xxx this one should be reachable)
 # C0302 Too many lines in module (would be nice to not count docstrings)
-# pylint: disable=R0914, R0904, C0302
+# pylint: disable=w0212, r0914, r0904, c0302
 
 # Historical note: we used to formally define the Schedulable type hint
 # but that ended up clobbering the documentation, and was more harmful
@@ -72,7 +73,8 @@ class PureScheduler:                                    # pylint: disable=r0902
         can be run simultaneously. None or 0 means no limit.
       timeout: can be an `int` or `float` and is expressed
        in seconds; it applies to the overall orchestration of that scheduler,
-       not to any individual job.
+       not to any individual job. Can be also ``None``, which means no timeout.
+      shutdown_timeout: same meaning as ``timeout``, but for the shutdown phase.
       watch: if the caller passes a :class:`~asynciojobs.watch.Watch`
         instance, it is used in debugging messages to show the time
         elapsed wrt that watch, instead of using the wall clock.
@@ -108,22 +110,25 @@ class PureScheduler:                                    # pylint: disable=r0902
 
     def __init__(self, *jobs_or_sequences,
                  jobs_window=None, timeout=None,
+                 shutdown_timeout=1,
                  watch=None, verbose=False):
 
-        self.jobs = set(Sequence._flatten(              # pylint: disable=W0212
-            jobs_or_sequences))
+        self.jobs = BestSet(Sequence._flatten(jobs_or_sequences))
         self.jobs_window = jobs_window
         # timeout is in seconds
         self.timeout = timeout
+        self.shutdown_timeout = shutdown_timeout
         self.watch = watch
         self.verbose = verbose
         # why does it fail ?
         # bool
         self._failed_critical = False
-        # False, or the intial timeout
+        # False, or the initial timeout
         self._failed_timeout = False
         # see also _record_beginning
         self._expiration = None
+        # avoid multiple shutdowns
+        self._did_shutdown = False
 
     # think of an scheduler as a set of jobs
     def update(self, jobs):
@@ -137,7 +142,7 @@ class PureScheduler:                                    # pylint: disable=r0902
         Returns:
           self: the scheduler object, for cascading insertions if needed.
         """
-        jobs = set(Sequence._flatten(jobs))             # pylint: disable=W0212
+        jobs = BestSet(Sequence._flatten(jobs))
         self.jobs.update(jobs)
         return self
 
@@ -150,9 +155,27 @@ class PureScheduler:                                    # pylint: disable=r0902
           job: a single ``Schedulable`` object.
 
         Returns:
-          self: the scheduler object, for cascading insertions if needed.
+          job: the job object, for convenience, as it can be needed to build
+            requirements later on in the script.
         """
         self.update([job])
+        return job
+
+    def remove(self, job):
+        """
+        Removes a single ``Schedulable`` object;
+        this method name is inspired from plain python ``set.remove()``
+
+        Parameters:
+          job: a single ``Schedulable`` object.
+
+        Raises:
+          KeyError: if job not in scheduler.
+
+        Returns:
+          self: the scheduler object, for cascading insertions if needed.
+        """
+        self.jobs.remove(job)
         return self
 
     def __len__(self):
@@ -197,7 +220,7 @@ class PureScheduler:                                    # pylint: disable=r0902
         return "FINE"
 
     ####################
-    def sanitize(self):
+    def sanitize(self, verbose=None):
         """
         This method ensures that the requirements relationship is closed within
         the scheduler. In other words, it removes any requirement attached to a
@@ -209,11 +232,25 @@ class PureScheduler:                                    # pylint: disable=r0902
         In any case it is crucial that this property holds
         for :meth:`co_run()` to perform properly.
 
+        Parameters:
+          verbose: if not None, defines verbosity for this operation.
+            Otherwise, the object's ``verbose`` attribute is used.
+            In verbose mode, jobs that are changed, i.e. that have
+            requirement(s) dropped because they are not part of
+            the same scheduler, are listed, together with their
+            container scheduler.
+
         Returns:
           bool: returns True if scheduler object was fine,
-          and False if at least one removal was needed.
+            and False if at least one removal was needed.
         """
 
+        if verbose is None:
+            verbose = self.verbose
+        if verbose:
+            container_label = self if not hasattr(self, 'label') \
+                else self.label                         # pylint: disable=e1101
+
         changes = False
         for job in self.jobs:
             before = len(job.required)
@@ -222,14 +259,17 @@ class PureScheduler:                                    # pylint: disable=r0902
             after = len(job.required)
             if before != after:
                 changes = True
-                if self.verbose:
-                    print(20 * '*',
-                          "WARNING: job {} has had {} requirements removed"
-                          .format(job, before - after))
+                if verbose:
+                    print(10 * '*',
+                          "WARNING: job {} in {} had {} requirements removed"
+                          .format(job, container_label, before - after))
+            # recursively scan nested schedulers
+            if isinstance(job, PureScheduler):
+                changes = job.sanitize(verbose) or changes
         return not changes
 
     ####################
-    def rain_check(self):
+    def check_cycles(self):
         """
         Performs a minimal sanity check.
         The purpose of this is primarily to check for cycles,
@@ -252,7 +292,7 @@ class PureScheduler:                                    # pylint: disable=r0902
             return True
         except Exception as exc:                        # pylint: disable=W0703
             if self.verbose:
-                print("rain_check failed", exc)
+                print("check_cycles failed", exc)
             return False
 
     ####################
@@ -363,34 +403,67 @@ class PureScheduler:                                    # pylint: disable=r0902
             result = "{" + result + "}"
         return result
 
-    def _exit_csv(self, discard_forever=True,
-                  compute_backlinks=False):
+    def _exit_csv(self, **exit_kwds):
+        'accepts same parameters as self.exit_jobs()'
         result = ", ".join(job.repr_id()
-                           for job in self.exit_jobs(
-                               discard_forever=discard_forever,
-                               compute_backlinks=compute_backlinks))
+                           for job in self.exit_jobs(**exit_kwds))
         if result:
             result = "{" + result + "}"
         return result
 
-    def some_entry_job(self):
-        """
-        Returns one randomly picked entry job
-        Needed when creating the dot format, because
-        of the specifics of that format; although it may be
-        that returning any job would do the job as well
-        """
-        for job in self.entry_jobs():
-            return job
-        raise ValueError("Internal error")
+    @staticmethod
+    def _middle_index(last):
+        return (last-1) // 2
 
-    def some_exit_job(self):
+    def _middle_entry_job(self):
         """
-        Same for exit nodes;
+        Returns an entry job; needed when creating the dot format, because
+        of the specifics of that format.
+
+        As an attempt to improve graphical layout,
+        we return the job that has its index in the middle of the entry jobs.
+        Also, we need to return an atomic job, not a scheduler/container.
         """
-        for job in self.exit_jobs():
-            return job
-        raise ValueError("Internal error")
+        # scan once
+        number_entries = sum(1 for _ in self.entry_jobs())
+        if not number_entries:
+            raise ValueError("no entry found")
+        # scan again until mid-way
+        entries = self.entry_jobs()
+        index = self._middle_index(number_entries)
+        for _, job in zip(range(index+1), entries):
+            pass
+        # pylint detects that job is possibly undefined,
+        # which could only occur if range(index+1) is empty
+        # but index is >= 0 so we're in the clear
+        candidate = job                                 # pylint: disable=w0631
+        if not isinstance(candidate, PureScheduler):
+            return candidate
+        return candidate._middle_entry_job()
+
+    def _middle_exit_job(self, **exit_kwds):
+        """
+        Same as ``_middle_entry_job``, for exit nodes;
+        accepts same parameters as ``self.exit_jobs()``
+        """
+        number_exits = sum(1 for _ in self.exit_jobs(**exit_kwds))
+        # no need to do this in any case from now on
+        exit_kwds['compute_backlinks'] = False
+        if not number_exits:
+            # second chance : allow for forever jobs
+            exit_kwds['discard_forever'] = False
+            number_exits = sum(1 for _ in self.exit_jobs(**exit_kwds))
+        if not number_exits:
+            raise ValueError("no exit found")
+        exits = self.exit_jobs(**exit_kwds)
+        index = self._middle_index(number_exits)
+        for _, job in zip(range(index+1), exits):
+            pass
+        # ditto
+        candidate = job                                 # pylint: disable=w0631
+        if not isinstance(candidate, PureScheduler):
+            return candidate
+        return candidate._middle_exit_job()
 
     def repr_entries(self):                             # pylint: disable=c0111
         return "entries={}".format(self._entry_csv())
@@ -419,7 +492,7 @@ class PureScheduler:                                    # pylint: disable=r0902
         as the reverse of Job.required
         """
         for job in self.jobs:
-            job._s_successors = set()                   # pylint: disable=W0212
+            job._s_successors = BestSet()               # pylint: disable=W0212
         for job in self.jobs:
             for req in job.required:
                 req._s_successors.add(job)              # pylint: disable=W0212
@@ -427,7 +500,7 @@ class PureScheduler:                                    # pylint: disable=r0902
     def _ensure_future(self, job, window, loop):        # pylint: disable=R0201
         """
         this is the hook that lets us make sure the created Task objects
-        have a backlink pointer to their corresponding job
+        have a backlink reference to their corresponding job
         """
         #
         # this is where we call co_run()
@@ -481,15 +554,11 @@ class PureScheduler:                                    # pylint: disable=r0902
         Similar but in order to clear the exceptions,
         we need to run gather() instead
         """
-        # do not use task._job.raised_exception()
-        # so we can use this with co_shutdown() tasks as well
-        # (these are not attached to a job)
-        exception_tasks = [task for task in tasks
-                           if task._exception]          # pylint: disable=W0212
+        exception_tasks = [task for task in tasks if task._exception]
         for task in exception_tasks:
             task.cancel()
-            # if debug is turned on, provide details on the exceptions
-            if debug:
+            # if DEBUG is turned on, provide details on the exceptions
+            if DEBUG:
                 job = task._job                         # pylint: disable=W0212
                 self._show_task_stack(
                     task,
@@ -519,34 +588,6 @@ class PureScheduler:                                    # pylint: disable=r0902
         print(sep, 'END ' + msg)
         print(sep)
 
-    async def co_shutdown(self):
-        """
-        Shut down the scheduler, by sending a message to all the jobs when
-        orchestration is over.
-
-        Typically for example, several jobs sharing the same ssh connection
-        will arrange for that connection to be kept alive across an entire
-        scheduler lifespan, but there is a need to tear these connections
-        down eventually.
-
-        Returns:
-          None
-        """
-
-        await self._feedback(None, "scheduler is shutting down...")
-        tasks = [asyncio.ensure_future(job.co_shutdown())
-                 for job in self.jobs]
-        # the done part is of no use here
-        _, pending = await asyncio.wait(
-            tasks, timeout=self._remaining_timeout())
-        if pending:
-            print("WARNING: {}/{} co_shutdown() methods"
-                  " have not returned within timeout"
-                  .format(len(pending), len(self.jobs)))
-            await self._tidy_tasks(pending)
-        # xxx should consume any exception as well ?
-        # self._tidy_tasks_exception(done)
-
     async def _feedback(self, jobs, state, force=False):
         """
         When self.verbose is set, provide feedback about the mentioned
@@ -562,38 +603,144 @@ class PureScheduler:                                    # pylint: disable=r0902
                 self.watch.print_elapsed()
             else:
                 Watch.print_wall_clock()
-        name = ""
-        # with this line we allow a PureScheduler to have a label
-        # it's mostly applicable to Scheduler instances, but can come in
-        # handy for pure Scheduler's as well
-        if hasattr(self, 'label'):
-            name = "({})".format(self.label)            # pylint: disable=e1101
+        name = self.stats()
 
         # general feedback when no job is specified by caller
         if jobs is None:
             print_time()
-            print("SCHEDULER{}: {}".format(name, state))
+            print("SCHEDULER {}: {}".format(name, state))
             return
-        if not isinstance(jobs, (list, set, tuple)):
+        if not isinstance(jobs, (list, BestSet, set, tuple)):
             jobs = (jobs,)
         for job in jobs:
             if not isinstance(job, AbstractJob):
                 # we expect a task here
                 job = job._job                          # pylint: disable=W0212
             print_time()
-            print("{} {}: {} {} {}"
+            print("{} {:8s}: {} {} {}"
                   .format(name, state,
                           job.repr_id(), job.repr_short(), job.repr_main()),
                   end="")
-            if self.verbose:
-                print(" {} {}"
-                      .format(job.repr_result(),
-                              job.repr_requires()),
-                      end="")
+            print(" {} {}"
+                  .format(job.repr_result(),
+                          job.repr_requires()),
+                  end="")
             print()
 
     ####################
-    def run(self, *args, loop=None, **kwds):
+    def shutdown(self):
+        """
+        A synchroneous wrapper around :meth:`co_shutdown()`.
+
+        Returns:
+          bool: True if everything went well, False otherwise;
+          see :meth:`co_shutdown()` for details.
+
+        """
+        loop = asyncio.get_event_loop()
+        return loop.run_until_complete(self.co_shutdown())
+
+    async def co_shutdown(self):
+        """
+        Shut down the scheduler, by sending the
+        :meth:`~asynciojobs.job.AbstractJob.co_shutdown()`
+        method to all the jobs, possibly nested.
+
+        Within nested schedulers, a job receives the `shutdown` event when its
+        **enclosing** scheduler terminates, and **not** at the end of the
+        **outermost** scheduler.
+
+        Also note that all job instances receive the 'co_shutdown()' method,
+        even the ones that have not yet started; it is up to the `co_shutdown()`
+        method to triage the jobs according to their life cycle status - see
+        :meth:`~asynciojobs.job.AbstractJob.is_running()` and similar.
+
+        This mechanism should be used only for minimal housekeeping only, it is
+        recommended that intrusive cleanup be made part of separate, explicit
+        methods.
+
+        :Note: typically in apssh for example, several jobs sharing the same ssh
+        connection need to arrange for that connection to be kept alive across
+        an entire scheduler lifespan, and closed later on. Historically there
+        had been an attempt to deal with this automagically, through the present
+        shutdown mechanism. However, this turned out to be the wrong choice, as
+        the choice of closing connections needs to be left to the user.
+        Additionally, with nested schedulers, this can become pretty awkward.
+        Closing ssh connections is now to be achieved explicitly through a call
+        to a specific apssh function.
+
+        Returns:
+          bool: True if all the
+           :meth:`~asynciojobs.job.AbstractJob.co_shutdown()`
+           methods attached to the jobs in the scheduler complete
+           within ``shutdown_timeout``, which is an attribute of the scheduler.
+           If the ``shutdown_timeout`` attribute on this object is ``None``,
+           no timeout is implemented.
+
+        Notes:
+          There is probably space for a lot of improvement here xxx:
+
+          - behaviour is unspecified if any of the co_shutdown()
+            methods raises an exception;
+          - right now, a subscheduler that sees a timeout expiration
+            does not cause the overall co_shutdown() to return ``False``,
+            which is arguable;
+          - another possible weakness in current implementation is that
+            it does not support to shutdown a scheduler that is still running.
+        """
+
+        # implementation note
+        # it could be tempting to have this code only send co_shutdown
+        # on pure jobs
+        # however this approach does not work in cases where a timeout occurs
+        # which causes a (sub-)nested scheduler to be kind of paused, without
+        # actually being aware of it (its co_run() code does not return at all)
+        # so it's much simpler to always scan all the whole scheduler tree,
+        # but to skip schedulers that have already shut down
+
+        if self._did_shutdown:
+            return
+
+        self._did_shutdown = True
+
+        tasks = [asyncio.ensure_future(job.co_shutdown())
+                 for job in self.jobs]
+
+        if not tasks:
+            return True
+
+        # warning: xxx this use a unique attribute to remember expiration
+        # so things might/probably will get messed up if one attempts
+        # to shutdown a scheduler while it is running.
+        self._record_beginning(self.shutdown_timeout)
+        timeout = self._remaining_timeout()
+
+        await self._feedback(None, "scheduler is shutting down...")
+
+        # the done part is of no use here
+        _, pending = await asyncio.wait(tasks, timeout=timeout)
+        # everything went fine
+        # NOTE however: here we say that sub-schedulers that expired in timeout
+        # should not impact the overall result; this is an arguable choice
+        if not pending:
+            return True
+
+        # with nested schedulers, this message would not be helpful
+        # because it is only guaranteed to show up at the toplevel
+        # and in addition the jobs count is local to the scheduler
+        # this is why this it's a verbose/feedback thing
+        await self._feedback(
+            None,
+            "WARNING: {}/{} co_shutdown() methods"
+            " have not returned within timeout"
+            .format(len(pending), len(self.jobs)))
+        await self._tidy_tasks(pending)
+        # we might need to consume any exception as well ?
+        # self._tidy_tasks_exception(done)
+        return False
+
+    ####################
+    def run(self, *args, **kwds):
         """
         A synchroneous wrapper around :meth:`co_run()`,
         please refer to that link for details on parameters and return value.
@@ -601,15 +748,13 @@ class PureScheduler:                                    # pylint: disable=r0902
         Also, the canonical name for this is ``run()`` but for historical
         reasons you can also use ``orchestrate()`` as an alias for ``run()``.
         """
-        if loop is None:
-            loop = asyncio.get_event_loop()
-        return loop.run_until_complete(
-            self.co_run(loop=loop, *args, **kwds))
+        loop = asyncio.get_event_loop()
+        return loop.run_until_complete(self.co_run(*args, **kwds))
 
     # define the alias for legacy
     orchestrate = run
 
-    async def co_run(self, loop=None):            # pylint: disable=R0912,R0915
+    async def co_run(self):                       # pylint: disable=R0912,R0915
 
         """
         The primary entry point for running a scheduler.
@@ -620,24 +765,18 @@ class PureScheduler:                                    # pylint: disable=r0902
 
         Proceeds to the end no matter what, except if either:
 
-        1. one critical job raises an exception, or
-
-        2. a timeout occurs.
+        * one critical job raises an exception, or
+        * a timeout occurs.
 
         Returns:
           bool: `True` if none of these 2 conditions occur, `False` otherwise.
 
-        Jobs marked as ``forever`` are not waited for. All jobs get
-        terminated through their `co_shutdown()` method.
-
-        Parameters:
-          loop: an asyncio event loop, it defaults to
-            ``asyncio.get_event_loop()``, which is almost certainly
-            the right value to use.
+        Jobs marked as ``forever`` are not waited for.
 
+        No automatic shutdown is performed, user needs to explicitly call
+        :meth:`co_shutdown()` or :meth:`shutdown()`.
         """
-        if loop is None:
-            loop = asyncio.get_event_loop()
+        loop = asyncio.get_event_loop()
         # create a Window no matter what; it will know what to do
         # also if jobs_window is None
         window = Window(self.jobs_window, loop)
@@ -688,8 +827,7 @@ class PureScheduler:                                    # pylint: disable=r0902
                                      timeout=self._remaining_timeout(),
                                      return_when=asyncio.FIRST_COMPLETED)
 
-            done_ok = {t for t in done
-                       if not t._exception}             # pylint: disable=W0212
+            done_ok = {t for t in done if not t._exception}
             await self._feedback(done_ok, "DONE")
             done_ko = done - done_ok
             await self._feedback(done_ko, "RAISED EXC.")
@@ -724,7 +862,7 @@ class PureScheduler:                                    # pylint: disable=r0902
                         done_job, "EXCEPTION occurred - on {}critical job"
                         .format("non-" if not done_job.is_critical() else ""))
                     # make sure these ones show up even if not in debug mode
-                    if debug:
+                    if DEBUG:
                         self._show_task_stack(done_task, "DEBUG")
             if critical_failure:
                 await self._tidy_tasks(pending)
@@ -738,13 +876,11 @@ class PureScheduler:                                    # pylint: disable=r0902
             # are we done ?
             # only account for not forever jobs (that may still finish, one
             # never knows)
-            done_jobs_not_forever = {
-                j for j in done
-                if not j._job.forever}                  # pylint: disable=W0212
+            done_jobs_not_forever = {j for j in done if not j._job.forever}
             nb_jobs_done += len(done_jobs_not_forever)
 
             if nb_jobs_done == nb_jobs_finite:
-                if debug:
+                if DEBUG:
                     print("PureScheduler.co_run: {} CLEANING UP at iter. {}/{}"
                           .format(4 * '-', nb_jobs_done, nb_jobs_finite))
                 if self.verbose and nb_jobs_forever != len(pending):
@@ -759,10 +895,10 @@ class PureScheduler:                                    # pylint: disable=r0902
             # go on : find out the jobs that can be added to the mix
             # only consider the ones that are right behind any of the the jobs
             # that just finished
+            # no need to use and BestSet here
             possible_next_jobs = set()
             for done_task in done:
-                possible_next_jobs.update(
-                    done_task._job._s_successors)       # pylint: disable=W0212
+                possible_next_jobs.update(done_task._job._s_successors)
 
             # find out which ones really can be added
             added = 0
@@ -795,8 +931,7 @@ class PureScheduler:                                    # pylint: disable=r0902
           int: total number of nodes in subject and nested schedulers
 
         """
-        return sum(job._job_count()                     # pylint: disable=w0212
-                   for job in self.jobs)
+        return sum(job._job_count() for job in self.jobs)
 
     def _set_sched_ids(self, start=1, id_format=None):
         """
@@ -822,42 +957,76 @@ class PureScheduler:                                    # pylint: disable=r0902
             i = job._set_sched_id(i, id_format)         # pylint: disable=w0212
         return i
 
-    ####################
+    def _set_sched_ids_safe(self, stack):
+        """
+        Similar to _set_sched_ids, in that the member jobs get their
+        internal sched_id set, for identifying jobs when expliciting
+        internal relationships in list_safe().
+
+        The difference here is that the numbering is not based on topological
+        order: jobs in a scheduler are scanned in random order, and jobs in
+        nested schedulers get prefixed.
+        """
+        root = ".".join(str(index) for index in stack)
+        for i, job in enumerate(self.jobs, 1):
+            job._sched_id = "{}.{}".format(root, i)     # pylint: disable=W0212
+            if isinstance(job, PureScheduler):
+                job._set_sched_ids_safe(stack+[i])      # pylint: disable=W0212
+
+    # ----
     def list(self, details=False):
         """
         Prints a complete list of jobs in topological order, with their status
         summarized with a few signs. See the README for examples and a legend.
 
-        Beware that this might raise an exception if :meth:`rain_check()`
+        Beware that this might raise an exception if :meth:`check_cycles()`
         would return ``False``, i.e. if the graph is not acyclic.
         """
         # so now we can refer to other jobs by their id when showing
         # requirements
         self._set_sched_ids()
         for job in self.topological_order():
-            job._list(details, 0)                       # pylint: disable=W0212
+            job._list(details, 0, True)                 # pylint: disable=W0212
 
     def list_safe(self):
         """
         Print jobs in no specific order, the advantage being that it
-        works even if scheduler is broken wrt :meth:`rain_check()`
+        works even if scheduler is broken wrt :meth:`check_cycles()`.
+        On the other hand, this method is not able to list requirements.
         """
-        for i, job in enumerate(self.jobs):
-            print(i, job)
+        self._set_sched_ids_safe([])
+        for job in self.jobs:
+            # pass as stack a list of indexes
+            job._list_safe(True)                        # pylint: disable=W0212
 
-    def __repr__(self):
-        # linter says unused variable but it is indeed used in f-string
-        nb_total = len(self.jobs)                       # pylint: disable=W0612
+
+    # ----
+    def _stats(self):
+        nb_total = len(self.jobs)
         done = {j for j in self.jobs if j.is_done()}
-        nb_done = len(done)                             # pylint: disable=W0612
+        nb_done = len(done)
         running = {j for j in self.jobs if j.is_running()}
         ongoing = running - done
-        nb_ongoing = len(ongoing)                       # pylint: disable=W0612
+        nb_ongoing = len(ongoing)
         idle = self.jobs - running
-        nb_idle = len(idle)                             # pylint: disable=W0612
-        return "{type} with {nb_done} done + {nb_ongoing} ongoing" \
-               " + {nb_idle} idle = {nb_total} job(s)" \
-            .format(type=type(self).__name__, **locals())
+        nb_idle = len(idle)
+        return (nb_done, nb_ongoing, nb_idle, nb_total)
+
+    def __repr__(self):
+        done, ongoing, idle, total = self._stats()
+        return ("{type} with {done} done + {ongoing} ongoing"
+                " + {idle} idle = {total} job(s)"
+                .format(type=type(self).__name__,
+                        done=done, ongoing=ongoing, idle=idle, total=total))
+
+    def stats(self):
+        """
+        Returns a string like e.g. ``2D + 3R + 4I = 9`` meaning that
+        the scheduler currently has 2 done, 3 running an 4 idle jobs
+        """
+        done, ongoing, idle, total = self._stats()
+        return ("{done}D + {ongoing}R + {idle}I = {total}"
+                .format(done=done, ongoing=ongoing, idle=idle, total=total))
 
     def debrief(self, details=False):
         """
@@ -912,15 +1081,30 @@ class PureScheduler:                                    # pylint: disable=r0902
                     continue
                 if not self.verbose:
                     print("non-critical: {}: exception {}"
-                          .format(j._get_text_label(),  # pylint: disable=W0212
-                                  j.raised_exception()))
+                          .format(j._get_text_label(), j.raised_exception()))
                     if self.verbose:
                         self._show_task_stack(
                             j, "non-critical job exception stack")
 
+    # ----
+    def iterate_jobs(self, scan_schedulers=False):
+        """
+        A generator that scans all jobs and subjobs
+
+        Parameters:
+          scan_schedulers: if set, nested schedulers are ignored,
+            only actual jobs are reported; otherwise, nested schedulers
+            are listed as well.
+        """
+        if scan_schedulers:
+            yield self
+        for job in self.jobs:
+            yield from job._iterate_jobs(scan_schedulers=scan_schedulers)
+
+    # ----
     # graphical outputs
-    #
-    # in a first attempt we had one function to store a dot format into a file
+
+    # in a first attempt we had one function to store a _ format into a file
     # and another one to build the graph natively; hence duplication of code
     # there are simpler means to do that
     # in addition with nested schedulers things become a bit messy, so
@@ -941,8 +1125,6 @@ class PureScheduler:                                    # pylint: disable=r0902
         See also `Wikipedia on DOT`_ for a list of tools that support the
         ``dot`` format.
 
-
-
         As a general rule, ``asynciojobs`` has a support for producing
         `DOT Format`_ but stops short of actually importing ``graphviz``
         that can be cumbersome to install, but for the notable exception
@@ -960,7 +1142,6 @@ class PureScheduler:                                    # pylint: disable=r0902
 DOT_%28graph_description_language%29
         """
         self._set_sched_ids()
-        # xxx should maybe use/show scheduler's label if set
         return "digraph asynciojobs" + self._dot_body(DotStyle())
 
     def _dot_body(self, dot_style):
@@ -994,7 +1175,7 @@ DOT_%28graph_description_language%29
 
                     # upstream is a scheduler
                     else:
-                        from_node = req.some_exit_job()
+                        from_node = req._middle_exit_job()
                         cluster_name = req.dot_cluster_name()
                         result += ("{} -> {} [ltail={}];\n"
                                    .format(from_node.repr_id(),
@@ -1003,13 +1184,11 @@ DOT_%28graph_description_language%29
 
             # nested scheduler
             else:
-                # xxx should maybe use/show scheduler's label if set
                 # insert a subgraph instead
 
                 cluster_name = job.dot_cluster_name()
                 result += "subgraph {}".format(cluster_name)
-                result += job._dot_body(                # pylint: disable=w0212
-                    job.dot_style())
+                result += job._dot_body(job.dot_style())
 
                 for req in job.required:
 
@@ -1017,15 +1196,15 @@ DOT_%28graph_description_language%29
                     if not isinstance(req, PureScheduler):
                         result += ("{} -> {} [lhead={}];\n"
                                    .format(req.repr_id(),
-                                           job.some_entry_job().repr_id(),
+                                           job._middle_entry_job().repr_id(),
                                            cluster_name))
 
                     # upstream is a scheduler as well
                     else:
                         src_cluster_name = req.dot_cluster_name()
                         result += ("{} -> {} [lhead={} ltail={}];\n"
-                                   .format(req.some_exit_job().repr_id(),
-                                           job.some_entry_job().repr_id(),
+                                   .format(req._middle_exit_job().repr_id(),
+                                           job._middle_entry_job().repr_id(),
                                            cluster_name,
                                            src_cluster_name))
 
@@ -1073,7 +1252,8 @@ DOT_%28graph_description_language%29
         for how this works.
 
         The dependency from ``asynciojobs`` to ``graphviz`` is limited
-        to this method, as it is the only place that needs it,
+        to this method and :meth:`export_as_pngfile()`,
+        as it these are the only places that need it,
         and as installing ``graphviz`` can be cumbersome.
 
         For example, on MacOS I had to do both::
@@ -1084,3 +1264,30 @@ DOT_%28graph_description_language%29
 
         from graphviz import Source
         return Source(source=self.dot_format())
+
+    def export_as_pngfile(self, filename):
+        """
+        Convenience wrapper that creates a png file. Like :meth:`graph()`,
+        it requires the ``graphviz`` package to be installed.
+
+        Parameters:
+          filename: output filename, without the ``.png`` extension
+        Returns:
+          created file name
+
+        Notes:
+          - This actually uses the binary `dot` program.
+          - A file named as the output but with a ``.dot`` extension
+            is created as an artefact by this method.
+        """
+        # we refrain from using graph.format / graph.render
+        # because with that method we cannot control the location
+        # of the .dot file; that is dangerous when using e.g.
+        #    scheduler.export_as_pngfile(__file__)
+        import os
+        dotfile = "{}.dot".format(filename)
+        pngfile = "{}.png".format(filename)
+        self.export_as_dotfile(dotfile)
+        os.system("dot -Tpng {} -o {}"
+                  .format(dotfile, pngfile))
+        return pngfile
